#include "ByteVoxmap.h"
#include <math.h>

static void check_(ByteVoxmap *this, ulong i, ulong j, ulong k);


void byte_voxmap_new(
    ByteVoxmap **p_this,
    ulong width,
    ulong height,
    ulong depth) {
  ByteVoxmap *this;

  if (width == 0) {
    throw("Width must be nonzero");
  }

  if (height== 0) {
    throw("Height must be nonzero");
  }

  if (depth== 0) {
    throw("Depth must be nonzero");
  }

  this = newo(p_this);

  this->width = width;
  this->height = height;
  this->depth = depth;
  this->numVoxels = width * height * depth;

  newt3(&this->_voxels, width, height, depth);
}

void byte_voxmap_new_from_double(
    ByteVoxmap **p_this,
    DoubleVoxmap *other) {
  byte ***this_voxels;
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double ***other_voxels;
  ByteVoxmap *this;

  width = other->width;
  height = other->height;
  depth = other->depth;

  byte_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  this_voxels = this->_voxels;
  other_voxels = other->_voxels;

  for (i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        this_voxels[i][j][k] = byte_bound(other_voxels[i][j][k] * 255);
      }
    }
  }
}

void byte_voxmap_new_from_double_diff(
    ByteVoxmap **p_this,
    DoubleVoxmap *diff) {
  byte ***out;
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double ***in;
  ByteVoxmap *this;

  width = diff->width;
  height = diff->height;
  depth = diff->depth;

  byte_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  in = diff->_voxels;
  out = this->_voxels;

  for (i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        out[i][j][k] = byte_encode_diff(in[i][j][k]);
      }
    }
  }
}

void byte_voxmap_new_from_abs_double_diff(
    ByteVoxmap **p_this,
    DoubleVoxmap *other) {
  byte ***this_voxels;
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double ***other_voxels;
  ByteVoxmap *this;

  width = other->width;
  height = other->height;
  depth = other->depth;

  byte_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  this_voxels = this->_voxels;
  other_voxels = other->_voxels;

  for (i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        this_voxels[i][j][k] = byte_bound(fabs(other_voxels[i][j][k]) * 255);
      }
    }
  }
}

void byte_voxmap_delete(ByteVoxmap **p_this) {
  ByteVoxmap *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletet3v(&this->_voxels, this->width, this->height);

  deleteo(p_this);
}

void byte_voxmap_print(ByteVoxmap *this, Printer *p) {
  ulong ii;
  ulong j;
  ulong k;

  sprint(p, "ByteVoxmap");
  rprint(p, "Width: %lu voxel%s", this->width, plural_s(this->width));
  rprint(p, "Height: %lu voxel%s", this->height, plural_s(this->height));
  rprint(p, "Depth: %lu voxel%s", this->depth, plural_s(this->depth));

  sprint(p, "Contents");

  for (k = 0; k < this->depth; k++) {
    sprint(p, "Slice %lu", k);

    for (j = 0; j < this->height; j++) {
      sprint(p, "Row %lu", j);

      for (ii = 0; ii < this->width; ii++) {
        rprint(p, "[%lu] 0x%02x", ii, byte_voxmap_get(this, ii, j, k));
      }

      fprint(p);
    }

    fprint(p);
  }

  fprint(p);
  fprint(p);
}

byte byte_voxmap_get(ByteVoxmap *this, ulong i, ulong j, ulong k) {
  check_(this, i, j, k);
  return this->_voxels[i][j][k];
}

void byte_voxmap_set(ByteVoxmap *this, ulong i, ulong j, ulong k, byte b) {
  check_(this, i, j, k);
  this->_voxels[i][j][k] = b;
}

void double_voxmap_new_from_byte(
    DoubleVoxmap **p_this,
    ByteVoxmap *other) {
  byte ***other_voxels;
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double ***this_voxels;
  DoubleVoxmap *this;

  width = other->width;
  height = other->height;
  depth = other->depth;

  double_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  this_voxels = this->_voxels;
  other_voxels = other->_voxels;

  for (i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        this_voxels[i][j][k] = other_voxels[i][j][k] / 255.0;
      }
    }
  }
}

void double_voxmap_new_from_byte_diff(
    DoubleVoxmap **p_this,
    ByteVoxmap *other) {
  byte ***other_voxels;
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double ***this_voxels;
  DoubleVoxmap *this;

  width = other->width;
  height = other->height;
  depth = other->depth;

  double_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  this_voxels = this->_voxels;
  other_voxels = other->_voxels;

  for (i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        this_voxels[i][j][k] = byte_decode_diff(other_voxels[i][j][k]);
      }
    }
  }
}


static void check_(ByteVoxmap *this, ulong i, ulong j, ulong k) {
  if (i >= this->width) {
    throw("i (%lu) must be less than width (%lu)", i, this->width);
  }

  if (j >= this->height) {
    throw("j (%lu) must be less than height (%lu)", j, this->height);
  }

  if (k >= this->depth) {
    throw("k (%lu) must be less than depth (%lu)", k, this->depth);
  }
}
