#include "pvl_nc_file.h"

#define MAX_BYTES_PER_SLAB_ (1024UL * 1024 * 1024)

static uint read_uint_(File *f);

static void write_uint_(File *f, uint u);


ByteVoxmap *byte_voxmap_new_load_pvl_nc(
    ByteVoxmap **p_this,
    char *filename_stub) {
  byte b;
  byte ***voxels;
  char *_filename = null;
  char *_element = null;
  char **_elements = null;
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong n;
  ulong num_slabs;
  ulong num_slices;
  ulong read_width;
  ulong read_height;
  ulong slab;
  ulong slab_size;
  ulong slice;
  ulong u;
  ulong v;
  ulong width;
  XmlFile *_xml = null;
  XmlFile *_xml2 = null;
  File *_f = null;
  ByteVoxmap *this;

  string_new_f(&_filename, "%s.pvl.nc", filename_stub);
  xml_file_new(&_xml, _filename, "r");
  string_delete(&_filename);

  xml_file_read_tag_attributes(_xml, "!DOCTYPE", "Drishti_Header");
  xml_file_new_memory_read_element(&_xml2, _xml, "PvlDotNcFileHeader");
  xml_file_delete(&_xml);
  xml_file_skip_element(_xml2, "rawfile");
  xml_file_read_element(_xml2, "voxeltype", &_element);

  if (string_equals(_element, "float")) {
    throw("Cannot handle float values");
  }

  string_delete(&_element);

  xml_file_read_element(_xml2, "gridsize", &_element);

  strings_new_split(&_elements, &n, _element, ' ');

  if (n != 3) {
    throw("Got %lu splits instead of 3 for gridsize", n);
  }

  string_delete(&_element);
  depth = parse_ulong(_elements[0]);
  height = parse_ulong(_elements[1]);
  width = parse_ulong(_elements[2]);
  delete(&_elements, 3, string_delete);

  xml_file_skip_element(_xml2, "voxelunit");
  xml_file_skip_element(_xml2, "voxelsize");
  xml_file_skip_element(_xml2, "description");

  xml_file_read_element(_xml2, "slabsize", &_element);
  slab_size = parse_ulong(_element);
  string_delete(&_element);

  xml_file_skip_element(_xml2, "rawmap");

  xml_file_read_element(_xml2, "pvlmap", &_element);
  strings_new_split(&_elements, &n, _element, ' ');
  string_delete(&_element);

  for (u = 0; u < n; u++) {
    v = parse_long(_elements[u]);

    if (v < 0 || v > 255) {
      throw("Value %ld is outside byte range", v);
    }
  }

  delete(&_elements, n, string_delete);
  xml_file_delete(&_xml2);

  num_slabs = (depth + slab_size - 1) / slab_size;

  byte_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  voxels = this->_voxels;

  for (k = 0, slab = 0; slab < num_slabs; slab++) {
    string_new_f(&_filename,  "%s.pvl.nc.%03lu", filename_stub, slab + 1);
    file_new(&_f, _filename, "r");

    if ((b = file_read_byte(_f)) != 0x00) {
      throw("First byte of '%s' was 0x%02x, not 0x00 as expected",
          _filename,
          b);
    }

    num_slices = read_uint_(_f);
    read_height = read_uint_(_f);
    read_width = read_uint_(_f);

    if (read_width != width) {
      throw("Read width %lu in file '%s' was not expected width %lu",
          read_width,
          _filename,
          width);
    }

    if (read_height != height) {
      throw("Read height %lu in file '%s' was not expected height %lu",
          read_height,
          _filename,
          height);
    }

    for (slice = 0; slice < num_slices; slice++, k++) {
      for (j = 0; j < height; j++) {
        for (i = 0; i < width; i++) {
          voxels[i][j][k] = file_read_byte(_f);
        }
      }
    }

    string_delete(&_filename);
    file_delete(&_f);
  }

  return this;
}

void byte_voxmap_save_pvl_nc(ByteVoxmap *this, char *filename_stub) {
  byte ***voxels;
  char *_filename = null;
  ulong i;
  ulong j;
  ulong k;
  ulong num_slabs;
  ulong num_slices;
  ulong slab;
  ulong slab_size;
  ulong slice;
  XmlFile *_xml = null;
  File *_f = null;

  slab_size = MAX_BYTES_PER_SLAB_ / (this->width * this->height);

  if (slab_size == 0) {
    throw("Each %lu x %lu slice of voxmap is too large for the .pvl.nc format",
        this->width,
        this->height);
  }

  num_slabs = (this->depth + slab_size - 1) / slab_size;

  string_new_f(&_filename, "%s.pvl.nc", filename_stub);
  xml_file_new(&_xml, _filename, "w");
  string_delete(&_filename);

  xml_file_write_token(_xml, "<!DOCTYPE Drishti_Header>\n");
  xml_file_write_token(_xml, "<PvlDotNcFileHeader>\n");
  xml_file_write_token(_xml, "  <rawfile></rawfile>\n");
  xml_file_write_token(_xml, "  <voxeltype>unsigned char</voxeltype>\n");
  xml_file_write_token(
      _xml,
      "  <gridsize>%lu %lu %lu</gridsize>\n",
      this->depth,
      this->height,
      this->width);
  xml_file_write_token(_xml, "  <voxelunit>no units</voxelunit>\n");
  xml_file_write_token(_xml, "  <voxelsize>1 1 1</voxelsize>\n");
  xml_file_write_token(
      _xml,
      "  <description>Created by John P. Costella</description>\n");
  xml_file_write_token(_xml, "  <slabsize>%lu</slabsize>\n", slab_size);
  xml_file_write_token(_xml, "  <rawmap>0 1</rawmap>\n");
  xml_file_write_token(_xml, "  <pvlmap>0 255 </pvlmap>\n");
  xml_file_write_token(_xml, "</PvlDotNcFileHeader>\n");
  xml_file_delete(&_xml);

  voxels = this->_voxels;

  for (slab = 0; slab < num_slabs; slab++) {
    string_new_f(&_filename,  "%s.pvl.nc.%03lu", filename_stub, slab + 1);
    file_new(&_f, _filename, "w");
    string_delete(&_filename);
    file_write_byte(_f, 0x00);
    num_slices = ((slab + 1) * slab_size > this->depth
        ? this->depth - slab * slab_size
        : slab_size);
    write_uint_(_f, num_slices);
    write_uint_(_f, this->height);
    write_uint_(_f, this->width);

    for (k = slab * slab_size, slice = 0; slice < num_slices; slice++, k++) {
      for (j = 0; j < this->height; j++) {
        for (i = 0; i < this->width; i++) {
          file_write_byte(_f, voxels[i][j][k]);
        }
      }
    }

    file_delete(&_f);
  }
}


static uint read_uint_(File *_f) {
  return (uint)file_read_byte(_f) | (uint)file_read_byte(_f) << 8
    | (uint)file_read_byte(_f) << 16 | (uint)file_read_byte(_f) << 24;
}

static void write_uint_(File *_f, uint u) {
  file_write_byte(_f, (byte)u);
  u >>= 8;
  file_write_byte(_f, (byte)u);
  u >>= 8;
  file_write_byte(_f, (byte)u);
  u >>= 8;
  file_write_byte(_f, (byte)u);
}
