#include "pvl_nc_top_down.h"

int main(int argc, char *argv[]) {
  byte a;
  byte orig_scale;
  byte v;
  sbyte scale;
  char *method;
  char *orig_stub;
  char *_stub = null;
  ulong depth;
  ulong dim;
  ulong height;
  ulong r;
  ulong width;
  ulong *_widths = null;
  ulong *_heights = null;
  ulong *_depths = null;
  double mad;
  Logger *_l = null;
  ByteVoxmap *_in = null;
  ByteVoxmap *_out = null;
  DoubleVoxmap *_approx = null;
  DoubleVoxmap *_diff = null;
  DoubleVoxmap *_final = null;
  DoubleVoxmap *_quantized_diff = null;
  DoubleVoxmap *_upsized = null;
  DoubleVoxmap *_orig = null;
  DoubleVoxmap *_recon = null;
  DoubleVoxmap **_reps = null;

  args_description(
      "Convert a .pvl.nc voxmap into a set of top-down .pvl.nc files, "
      "using the JPEG-Clear philosophy");
  args_string("filename-stub", "Base filename (without .pvl.nc)");
  args_optional_integer(
      "mks-generation",
      'a',
      3,
      "Generation of Magic Kernel Sharp to be used for resizing");
  args_optional_integer(
      "sharp-approximation",
      'v',
      3,
      "Approximation of the Sharp kernel to be used for resizing");
  args_parse(argc, argv);

  logger_new(&_l, stdout);

  orig_stub = args_get_string("filename-stub");
  a = (byte)args_get_bounded_integer("mks-generation", 1, 6);
  v = (byte)args_get_bounded_integer("sharp-approximation", 1, 7);
  method = MAGIC_RESIZER_MAGIC_KERNEL_SHARP;

  slog(_l, "Loading %s.pvl.nc", orig_stub);
  byte_voxmap_new_load_pvl_nc(&_in, orig_stub);
  flog(_l);

  slog(_l, "Converting byte voxmap to double");
  double_voxmap_new_from_byte(&_orig, _in);
  byte_voxmap_delete(&_in);
  flog(_l);

  dim = maximum(maximum(_orig->width, _orig->height), _orig->depth);

  for (orig_scale = 0; dim > 1; orig_scale++, dim++, dim >>= 1) {
  }

  new(&_reps, orig_scale + 1);
  donate(&_orig, &_reps[orig_scale]);
  _orig = _reps[orig_scale];

  new(&_widths, orig_scale + 1);
  new(&_heights, orig_scale + 1);
  new(&_depths, orig_scale + 1);

  scale = (sbyte)orig_scale;
  width = _orig->width;
  height = _orig->height;
  depth = _orig->depth;

  do {
    _widths[scale] = width;
    _heights[scale] = height;
    _depths[scale] = depth;

    scale--;
    width++;
    width /= 2;
    height++;
    height /= 2;
    depth++;
    depth /= 2;
  } while (scale >= 0);

  for (scale = (sbyte)orig_scale - 1, r = 2; scale >= 0; scale--, r *= 2) {
    slog(_l, "Downsizing to scale %d", scale);
    double_voxmap_new_resize(
        &_reps[scale],
        _orig,
        _widths[scale],
        _heights[scale],
        _depths[scale],
        1.0 / r,
        method,
        a,
        v);
    flog(_l);

    slog(_l, "Conversing double voxmap to byte");
    byte_voxmap_new_from_double(&_out, _reps[scale]);
    flog(_l);

    string_new_f(&_stub, "%s.%d", orig_stub, scale);
    slog(_l, "Saving representation at scale %d as %s", scale, _stub);
    byte_voxmap_save_pvl_nc(_out, _stub);
    byte_voxmap_delete(&_out);
    flog(_l);
    string_delete(&_stub);
  }

  slog(_l, "Converting the fundamental voxel to byte");
  byte_voxmap_new_from_double(&_out, _reps[0]);
  flog(_l);

  slog(_l, "Converting the byte fundamental voxel to double");
  double_voxmap_new_from_byte(&_approx, _out);
  byte_voxmap_delete(&_out);
  flog(_l);

  for (scale = 1; scale <= orig_scale; scale++) {
    slog(_l, "Upsizing to scale %d", scale);
    double_voxmap_new_resize(
        &_upsized,
        _approx,
        _widths[scale],
        _heights[scale],
        _depths[scale],
        2,
        method,
        a,
        v);
    double_voxmap_delete(&_approx);
    flog(_l);

    slog(_l, "Converting double voxmap to byte");
    byte_voxmap_new_from_double(&_out, _upsized);
    flog(_l);

    string_new_f(&_stub, "%s.upsized.%d", orig_stub, scale);
    slog(_l, "Saving upsized at scale %d as %s", scale, _stub);
    byte_voxmap_save_pvl_nc(_out, _stub);
    byte_voxmap_delete(&_out);
    flog(_l);
    string_delete(&_stub);

    slog(_l, "Computing the diff at scale %d", scale);
    double_voxmap_new_subtract(&_diff, _reps[scale], _upsized);
    flog(_l);

    slog(_l, "Converting double diff voxmap to byte diff encoding");
    byte_voxmap_new_from_double_diff(&_out, _diff);
    flog(_l);

    string_new_f(&_stub, "%s.diff.%d", orig_stub, scale);
    slog(_l, "Saving diff at scale %d as %s", scale, _stub);
    byte_voxmap_save_pvl_nc(_out, _stub);
    flog(_l);
    string_delete(&_stub);

    slog(_l, "Converting byte diff back to double");
    double_voxmap_new_from_byte_diff(&_quantized_diff, _out);
    byte_voxmap_delete(&_out);
    flog(_l);

    slog(_l, "Converting abs of double diff voxmap to byte");
    byte_voxmap_new_from_abs_double_diff(&_out, _diff);
    double_voxmap_delete(&_diff);
    flog(_l);

    string_new_f(&_stub, "%s.absdiff.%d", orig_stub, scale);
    slog(_l, "Saving abs-diff at scale %d as %s", scale, _stub);
    byte_voxmap_save_pvl_nc(_out, _stub);
    byte_voxmap_delete(&_out);
    flog(_l);
    string_delete(&_stub);

    slog(_l, "Computing the approximation at scale %d", scale);
    double_voxmap_new_add(&_approx, _upsized, _quantized_diff);
    double_voxmap_delete(&_upsized);
    double_voxmap_delete(&_quantized_diff);
    flog(_l);

    if (scale == orig_scale) {
      slog(_l, "Computing the final diff");
      double_voxmap_new_subtract(&_diff, _reps[scale], _approx);
      double_voxmap_delete(&_approx);
      flog(_l);

      slog(_l, "Converting final double diff to byte diff encoding");
      byte_voxmap_new_from_double_diff(&_out, _diff);
      double_voxmap_delete(&_diff);
      flog(_l);

      string_new_f(&_stub, "%s.diff.final", orig_stub);
      slog(_l, "Saving final diff as %s", _stub);
      byte_voxmap_save_pvl_nc(_out, _stub);
      byte_voxmap_delete(&_out);
      flog(_l);
      string_delete(&_stub);
    }
  }

  slog(_l, "Validation phase");

  string_new_f(&_stub, "%s.0", orig_stub);
  slog(_l, "Loading the fundamental voxel from %s", _stub);
  byte_voxmap_new_load_pvl_nc(&_in, _stub);
  flog(_l);
  string_delete(&_stub);

  slog(_l, "Converting byte voxmap to double");
  double_voxmap_new_from_byte(&_recon, _in);
  byte_voxmap_delete(&_in);
  flog(_l);

  if ((mad = double_voxmap_max_abs_diff(_recon, _reps[0])) >= 0.5 / 255) {
    throw("Fundamental voxmap incorrect: max abs diff %.16g", mad);
  }

  for (scale = 1; scale <= orig_scale; scale++) {
    slog(_l, "Upsizing to scale %d", scale);
    double_voxmap_new_resize(
        &_upsized,
        _recon,
        _widths[scale],
        _heights[scale],
        _depths[scale],
        2,
        method,
        a,
        v);
    double_voxmap_delete(&_recon);
    flog(_l);

    slog(_l, "Loading diff file at scale %d", scale);
    string_new_f(&_stub, "%s.diff.%d", orig_stub, scale);
    byte_voxmap_new_load_pvl_nc(&_in, _stub);
    string_delete(&_stub);
    flog(_l);

    slog(_l, "Converting byte diff back to double");
    double_voxmap_new_from_byte_diff(&_quantized_diff, _in);
    byte_voxmap_delete(&_in);
    flog(_l);

    slog(_l, "Adding in diff data at scale %d", scale);
    double_voxmap_new_add(&_recon, _upsized, _quantized_diff);
    double_voxmap_delete(&_upsized);
    double_voxmap_delete(&_quantized_diff);
    flog(_l);

    if ((mad = double_voxmap_max_abs_diff(_recon, _reps[scale])) > 3.0 / 255) {
      throw("Incorrect reconstruction at scale %d: "
          "max abs diff %.16g, byte %.16g",
          scale,
          mad,
          mad * 255);
    }
  }

  string_new_f(&_stub, "%s.diff.final", orig_stub);
  slog(_l, "Loading final diff from %s", _stub);
  byte_voxmap_new_load_pvl_nc(&_in, _stub);
  flog(_l);
  string_delete(&_stub);

  slog(_l, "Converting byte diff back to double");
  double_voxmap_new_from_byte_diff(&_quantized_diff, _in);
  byte_voxmap_delete(&_in);
  flog(_l);

  slog(_l, "Adding in final diff");
  double_voxmap_new_add(&_final, _recon, _quantized_diff);
  double_voxmap_delete(&_recon);
  double_voxmap_delete(&_quantized_diff);
  flog(_l);

  slog(_l, "Checking final reconstruction against original");

  if ((mad = double_voxmap_max_abs_diff(_final, _reps[orig_scale])) >= 0.5) {
    throw("Final reconstruction incorrect: max abs diff %.16g, byte %.16g",
        mad,
        mad * 255);
  }

  double_voxmap_delete(&_final);
  flog(_l);

  flog(_l);

  deletev(&_depths);
  deletev(&_heights);
  deletev(&_widths);

  delete(&_reps, orig_scale + 1, double_voxmap_delete);

  logger_delete(&_l);

  return 0;
}
